-------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   Clp.hs
--
-- Author:
--   Jan Hoffmann (2011)
--
-- Description: 
--   A CLP front end.
--   See http://projects.coin-or.org/Clp
--
--   RAML expects the binary clp to be in the path.
--   You can give the path to the binary by changing the variable clpBin.
--
--   The module provides two functions.
--
--   The function solve takes a list of constraints and a objective function
--   (and a path for the temp files), writes the constraints and the objective
--   function to an mps file, runs cpl on the file, and parses the solution
--   into a variable mapping (or throws an error if the constraint set could
--   not be solved).
--
--   The function cplResult is a parser for the lp_solve output. 


module Clp(solve) where

import qualified System.Cmd as Cmd     -- to execute shell commands
import qualified Data.Map as M
import Text.ParserCombinators.Parsec hiding (State)   -- to parse the output
import System.IO                       -- to work with files


import RAMLTypes hiding (Eq)
import Util


-- The path to the clp binary.
clpBin = "clp"


solve :: (Monad err) => Constraints -> [Objective] -> FilePath -> IO (err (M.Map CVarId Double))
-- The function solve takes a list of constraints and a objective function
-- (and a path for the temp files), writes the constraints and the objective
-- function to a file, runs lp_solve on the file, and parses the solution
-- into a variable mapping (or throws an error if the constraint set could
-- not be solved).

-- Haskell provides a nice way to create temp files.  We create
-- two in the temp dir: constrXXX.lp and solutionXXX.lp. Where XXX is a
-- random number.  We only need the file name and close the handles right away.

-- Then store the linar program, run lp_solve and store the output, 
-- read the solution from the output file, and parse the solution.
solve constraints objectives tempDir =
    do { (constrFileName, constrFile) <- openTempFile tempDir "constr.lp"  
       ; (solutionFileName, solutionFile) <- openTempFile tempDir "solution.clp"
       ; hClose solutionFile
       ; hSetBuffering constrFile (BlockBuffering Nothing)
       ; lpOutput constrFile constraints objectives
       ; hClose constrFile                                                       
       ; Cmd.system $ clpBin ++ " " ++ constrFileName ++ " -solv -solution " ++ solutionFileName ++ " > /dev/null"
       ; solutionString <- readFile solutionFileName                                         
       ; let solution = case parse clpResult solutionFileName solutionString of
                           Left err -> fail ("Clp finished unsuccessful. " ++ (show err))
                           Right x -> return x
       ; (Cmd.system $ "rm " ++ constrFileName ++ " " ++ solutionFileName)
       ; return solution
       }


lpOutput file constraints objectives =
    do { hPutStr file "Minimize\n"
       ; sequence_ [ hPutStr file (" + "++(show c)) | c <- objectives ]
       ; hPutStr file "\n"
       ; hPutStr file "Subject To\n"
       ; sequence_ [ hPutStr file (normalizeConstr c) | c <- constraints ]
       ; hPutStr file "End"
       }


normalizeConstr constraint =
    case constraint of
       CLeq lhs rhs -> let (lhs',q) = normalize lhs rhs in
                       (show $ CLeq lhs' [CConst q]) ++ "\n"
       CGeq lhs rhs -> let (lhs',q) = normalize lhs rhs in
                       (show $ CGeq lhs' [CConst q]) ++ "\n"
       CEq lhs rhs  -> let (lhs',q) = normalize lhs rhs in
                       (show $ CEq lhs' [CConst q]) ++ "\n"
  where
    normalize lhs rhs = 
      let (varsL,qL) = foldl processLhs ([],0) lhs
          (varsR,qR) = foldl processRhs ([],0) rhs
      in  (varsL++varsR,qL+qR)

    processLhs (vars,q) constr = case constr of
      CConst c -> (vars,q-c)
      var -> (var:vars,q)

    processRhs (vars,q) constr = case constr of
      CVar x -> (CnVar (-1) x:vars,q)
      CnVar n x -> (CnVar (-n) x:vars,q)
      CConst c -> (vars,q+c)


-- The parser for the lp_solve output.
-- The output can start with some empty lines.
-- Then there is ether the line "This problem is infeasible" (unsolvable)
-- or the line "Actual values of the variables:" (variableValues).
-- In the first we raise an error.
-- In the second case we collect the values of the variables.
clpResult = unsolvable
            <|> do { string "optimal"
                   ; spaces
                   ; skipLine
                   ; variableValues
                   }
            <|> do { fail "Something went wrong with the Clp output."
                   }
   where
      unsolvable = do { string "infeasible"
                      ; fail $ "The linear program is infeasible."
                      }
      
      variableValues = do { map <- many varValue
                          ; return $ M.fromList map
                          }
      
      varValue  = do { spaces
                     ; many digit
                     ; spaces
                     ; varId <- varName
                     ; spaces
                     ; value <- double
                     ; skipLine
                     ; return (varId,value)
                     }
      
      
      varName = do { l  <- letter -- Variable Names are much simpler.  But why not allow more?
                   ; ls <- many ( alphaNum <|> oneOf "_[]{}/.&#$%~'@^" )
                   ; return $ l:ls
                   }


      double = do { sign <- do { char '-'
                               ; return "-"
                               }
                         <|> do { char '+'
                                ; return "+"
                                }
                         <|> return ""
                  ; ds <- many digit
                  ; do { char '.' 
                       ; es <- many digit
                       ; return $ (read (sign ++ ds ++ "."++ es) :: Double)
                       }
                    <|> (return $ ((read $ sign ++ ds) :: Double))
                  }

      
      skipLine = do newline
                    return ()
                    <|> do anyChar
                           skipLine
      
      
