--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   MBinomial.hs
--
-- Author:
--   Klaus Aehlig (2010), Jan Hoffmann (2011)
--
-- Description:
--   This module contains datastructures for the representation of
--     - binomial coefficients with a variable (like "n over 2") and
--     - multivariate rational polynomials that are given in terms
--       of the standard basis.

module MBinomial where

import MultivariatePol
import RAMLTypes
import qualified Data.Map as M
import qualified Data.List as L

data Polynomial = Polynomial [VarId] (M.Map [Integer] Rational)

appendWithSep sep [t] = t
appendWithSep sep (t:ts) = t ++ sep ++ appendWithSep sep ts

showmonomial vars key ratcoeff = 
    let (sign,coeff) = if ratcoeff < 0 then ("-", fromRational ((-1) * ratcoeff)) else ("+", fromRational ratcoeff)
        varsexp = zip vars key
        factors = varsexp >>= \(x,e) -> if e == 0 then fail "1" else if e==1 then return x else return  (x ++ "^" ++ show e)
        sMonomial = if null factors then show coeff else if coeff == 1 then appendWithSep "*" factors else show coeff ++ "*" ++ appendWithSep "*" factors
    in (sign,sMonomial)

                    
comparelex []     [] = EQ
comparelex []     _  = LT
comparelex _      [] = GT
comparelex (a:as) (b:bs) = if a > b then GT else if b > a then LT else comparelex as bs

appendWithSign [("+",t)]  = t
appendWithSign [(s,t)]    =  s ++ t
appendWithSign ((s,t):ts) = appendWithSign ts ++ " " ++ s ++ " " ++ t

showpoly (Polynomial vars f) =
    let summands = L.sortBy comparelex (M.keys f) >>= \k -> let c = f M.! k in if c == 0 then fail "0" else return $ showmonomial vars k c
    in if null summands then "0" else appendWithSign summands

instance Show Polynomial where show = showpoly


-- Multiplying two Polynomials, with the same variables

multpol f g = 
    foldl (M.unionWith (+))  M.empty
    (combine [(M.toList f),(M.toList g)]
     >>= \ [(vf,cf),(vg,cg)] -> return $ M.fromList [(zipWith (+) vf vg, (f M.! vf) * (g M.! vg))])

-- filling up missing variables with exponent 0; vars must be a subset of vars'

castexp vars vars' key =
    let varskeys = zip vars key
    in map (\x -> case (lookup x varskeys) of
                  Nothing  -> 0
                  Just c   -> c) vars'

-- basic operations on Polynomials

multp (Polynomial vars f)(Polynomial vars' g) =
    let newvars = L.union vars vars'
        newf = M.mapKeys (castexp vars  newvars) f
        newg = M.mapKeys (castexp vars' newvars) g
    in Polynomial newvars (multpol newf newg)

addp (Polynomial vars f)(Polynomial vars' g) =
    let newvars = L.union vars vars'
        newf = M.mapKeys (castexp vars  newvars) f
        newg = M.mapKeys (castexp vars' newvars) g
    in Polynomial newvars (M.unionWith (+) newf newg)

subp f (Polynomial vars g) =
    f `addp` (Polynomial vars (M.map ((*) (-1)) g))

constpol c = Polynomial [] (M.fromList [([],c)])

varp var = Polynomial [var] (M.fromList [([1],1)])

-- binomial coefficients

binom x k = foldl multp (constpol 1) [varp x `subp` constpol i | i<- [0..k-1]]
            `multp` constpol (1 / fac k)

fac n | n<1 = 1
fac n = n * fac (n-1)


-- Naming scheme for the variables

data VName = VNUnit 
           | VNNat VarId
           | VNList VarId VName 
           | VNTree VarId VName 
           | VNTuple [VName] deriving (Eq,Show)

-- Transforming TIndices into standard polynomials, using a given naming scheme

toPolynomial :: VName -> TIndices -> Polynomial

toPolynomial _              TIUnit      = constpol 1
toPolynomial (VNNat x)     (TIList is)  = binom x (toRational (length is))
toPolynomial (VNTuple vns) (TITuple is) = foldl multp (constpol 1) (zipWith toPolynomial vns is)
toPolynomial (VNTree x vn) (TIList is)  = foldl multp (constpol 1) (map (toPolynomial vn) is)
                                          `multp` binom x (toRational (length is))
toPolynomial (VNList x vn) (TIList is)  = foldl multp (constpol 1) (map (toPolynomial vn) is)
                                          `multp` binom x (toRational (length is))




-- explaining what the individual variables mean

ordinal 0 = "first"
ordinal 1 = "second"
ordinal 2 = "third"
ordinal n = show (n+1) ++ "'th"


explain VNUnit        place  = []
explain (VNNat  x   ) place  = [x ++ " is the value " ++ place]
explain (VNList x vn) place  = [x ++ " is the length " ++ place] `L.union`  explain vn ("of the elements " ++ place)
explain (VNTree x vn) place  = [x ++ " is the size " ++ place] `L.union`  explain vn ("of the nodes " ++ place)
explain (VNTuple vns) place  = [0..(length vns) -1] >>= \i -> explain (vns !! i) ("of the " ++ ordinal i ++ " component " ++ place)

-- given a type and some variables assign them to all possible places (and return the unsued variables)

nametype (x:xs) (TNat)   = (VNNat x,xs)
nametype (x:xs) (TList t)   = let (vt,xs') = nametype xs t in (VNList x vt,xs')
nametype (x:xs) (TBTree t)   = let (vt,xs') = nametype xs t in (VNTree x vt,xs')
nametype vars   (TTuple ts) = foldl (\(VNTuple vts,xs) -> \t -> let (vt,xs') = nametype xs t in (VNTuple (vts++ [vt]),xs'))
                                    (VNTuple [],vars)
                                    ts
nametype vars   _           = (VNUnit, vars) -- atomic types


-- describe, given a bunch of variables and a type, a Multivariate polynomial

describePol vars t f =
    let (vt,_) = nametype vars t
        pol  = foldl addp (constpol 0) (map (\ i -> (toPolynomial vt i) `multp` constpol (f M.! i)) (M.keys f))
        explanations = explain vt "of the input"
    in appendWithSep "\n" ([show pol] ++ if null explanations then [] else ["where"] ++ map ((++) "   ") explanations)


-- evaluate a polynomial (and get a rational)

evalPoly :: Polynomial -> M.Map VarId Rational -> Rational

evalPoly (Polynomial vars pol) valMap =  foldl evalMonom 0  (M.toList pol)
  where
    vals = map (\x -> valMap M.! x) vars

    evalMonom acc (is,q) = q * (foldl (\acc (v,i) -> acc*(v^i)) 1 (zip vals is)) + acc


-- need evaluation that results in an int

evalBinom _ 0 = 1
evalBinom 0 _ = 0
evalBinom n k = (fac n) `div` ((fac (n-k))*(fac k) )

