{-# OPTIONS_GHC  -XTypeSynonymInstances -XFlexibleInstances  #-}

--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   RAMLTypes.hs
--
-- Author:
--   Jan Hoffmann (2009)
--
-- Description:
--   The datastructures of the prototype together with basic function
--   such as show.
--
--   The file is divided into the following sections:
--    - Expressions
--    - Simple Types and Programs
--    - Parsed Objects
--    - Constraints
--    - Resource Metrics
--    - Values and Sizes
--    - Misc
--
--   If something deserves a comment the comment is made in the
--   corresponding section.



module RAMLTypes where

import Data.List
import qualified Data.Map as M
import Text.ParserCombinators.Parsec (SourcePos)

import Util


------------------------------------------------------
-- Expressions
------------------------------------------------------

type VarId = String

type FunId = String

data MatchType = Dest | NonDest
   deriving (Eq,Show)

data Kind = Free | Normal
   deriving (Eq,Show)


-- Expressions and Declarations are "Parsed" and have therefore
-- a function sourePos that shows their position in the input file:
data Exp = 
           -- Constants
           EUnit SourcePos
         | ETrue SourcePos
         | EFalse SourcePos
         | ENat Int SourcePos

           -- Variables
         | EVar VarId SourcePos

         -- Functions
         | EApp FunId Exp SourcePos
         | EUnaryOp UnaryOp Exp SourcePos
         | EBinOp BinOp Exp Exp SourcePos

         -- Natural Numbers
         | ESucc Exp SourcePos
         | EToInt Exp SourcePos
         | EMatchN Exp Exp VarId Exp SourcePos         

           -- Let / Conditional
         | ELet VarId Exp Exp SourcePos
         | ECond Exp Exp Exp SourcePos

           -- Lists
         | ENil SourcePos
         | ECons Exp Exp SourcePos
         | EMatchL MatchType Exp Exp VarId VarId Exp SourcePos

           -- Tuples
         | ETuple [Exp] SourcePos
         | EMatchT Exp [VarId] Exp SourcePos

           -- Binary Trees
         | ELeaf SourcePos
         | ENodeB Exp Exp Exp SourcePos
         | EMatchB MatchType Exp Exp VarId VarId VarId Exp SourcePos

          -- Ticks
         | ETick Double SourcePos

         deriving (Eq)


instance Show Exp where
    show e = showTabs e 0
     where
       showTabs expr tabs = case expr of
            EUnit _ -> "()"

            ETrue _ -> "true"

            EFalse _ -> "false"

            ENat n _ ->  show n

            EApp f e _ -> f ++ " " ++ (show e)

            EVar x _ -> x

            EUnaryOp op e _ -> "(" ++ (show op) ++ "(" ++ (show e) ++ "))"

            EBinOp op e1 e2 _ -> "(" ++ (show e1) ++ ")" ++ (show op) ++ "("++(show e2)++")"

            ESucc e _ -> "S " ++ (show e)

            EToInt e _ -> "toInt " ++ (show e)

            EMatchN e e1 x e2 _ ->
                          "match " ++ (show e) ++ " with" ++
                                   "\n" ++ (mkSpaces tabs) ++ "0 -> " ++ (showTabs e1 (tabs+7)) ++
                                   "\n" ++ (mkSpaces tabs) ++ "S("++x++") -> " 
                                        ++ (showTabs e2 (tabs+(length x)+8))

            ELet x e1 e2 _ -> "let " ++ x ++ " = " ++ (showTabs e1 ((length x)+7) ) ++" in" ++
                               "\n" ++ (mkSpaces tabs) ++ (showTabs e2 tabs)

            ECond e e1 e2 _ -> "if " ++ (show e) ++ 
                               "\n"++(mkSpaces tabs)++"then "++(showTabs e1 (tabs+5))++
                               "\n"++(mkSpaces tabs)++"else "++(showTabs e2 (tabs+5))

            ENil _ -> "nil"

            ECons e1 e2 _ -> "cons(" ++ (show e1) ++ "," ++ (show e2) ++ ")"

            EMatchL matchType e e1 x_h x_t e2 _ ->
                                let maybeD = case matchType of
                                               Dest -> "D"
                                               NonDest -> ""
                                in "match" ++ maybeD ++ " " ++ (show e) ++ " with" ++
                                   "\n" ++ (mkSpaces tabs) ++ "nil -> " ++ (showTabs e1 (tabs+7)) ++
                                   "\n" ++ (mkSpaces tabs) ++ "("++x_h++"::"++x_t++") -> " 
                                        ++ (showTabs e2 (tabs+(length (x_h++x_t)+8)))

            ETuple es _ -> "(" ++ (showWithSep "," es) ++")"

            EMatchT e xs e1 _ -> let tuple = "("++ (showWithSep "," xs) ++") -> " in
                                   "match " ++ (show e) ++ " with" ++
                                   "\n" ++ (mkSpaces tabs) ++ tuple ++ (showTabs e1 (tabs+(length tuple)))

            ELeaf _ -> "leaf"

            ENodeB e1 e2 e3 _ -> "node(" ++ (show e1) ++ "," ++ (show e2) ++ "," ++ (show e3) ++ ")"

            EMatchB matchType e e1 v t1 t2 e2 _ ->
                                let maybeD = case matchType of
                                               Dest -> "D"
                                               NonDest -> ""
                                in "match" ++ maybeD ++ " " ++ (show e) ++ " with" ++
                                   "\n" ++ (mkSpaces tabs) ++ "leaf -> " ++ (showTabs e1 (tabs+7)) ++
                                   "\n" ++ (mkSpaces tabs) ++ "node("++v++","++t1++","++t2++") -> " 
                                        ++ (showTabs e2 (tabs+(length (v++t1++t2)+10)))

            ETick q _ -> "tick(" ++ (show q) ++ ")"


mkSpaces n = concat [ " " | _ <- [1..n] ]



data BinOp =   Add
             | Sub
             | Mult
             | Mod
             | Div
             | And
             | Or
             | Eq
             | LessEq
             | GreaterEq
             | Less
             | Greater
            deriving (Eq)

instance Show BinOp where

   show op = case op of
                Add -> "+"
                Sub -> "-"
                Mult -> "*"
                Mod -> "mod"
                Div -> "div"
                And -> "and"
                Or -> "or"
                Eq -> "=="
                LessEq -> "<="
                GreaterEq -> ">=" 
                Less -> "<"
                Greater -> ">" 


data UnaryOp =   UNot
               | UMinus
               | UPlus
              deriving (Eq)


instance Show UnaryOp where

   show op = case op of
                UNot -> "not"
                UMinus -> "-"
                UPlus -> "+"



------------------------------------------------------
-- Simple Types and Programs
------------------------------------------------------


data Type = TInt
          | TNat
          | TNatConst
          | TBool
          | TUnit
          | TNil
          | TList Type
          | TTuple [Type]
          | TLeaf
          | TBTree Type
         deriving (Eq,Ord)

instance Show Type where

   show tt = case tt of
               TInt -> "int"
               TNat -> "nat"
               TNatConst -> "number"
               TBool -> "bool"
               TUnit -> "unit"
               TNil -> "L(A)"
               TLeaf -> "T(A)"
               TList t -> case t of
                            TTuple _ -> "L" ++ (show t)
                            _ -> "L(" ++ (show t) ++ ")"
               TTuple ts -> "(" ++ (showWithSep "," ts) ++ ")"
               TBTree t -> case t of
                             TTuple _ -> "T" ++ (show t)
                             _ -> "T(" ++ (show t) ++ ")"


class Typed a where
    typeOf :: a -> Type

instance Typed Type where
    typeOf = id


data FunType = FunType Type Type
             deriving (Show,Eq)

data Dec = TypeDec FunId FunType SourcePos
         | FunDec FunId [VarId] Exp SourcePos
         deriving (Show,Eq)


type Program = ([Dec],Exp)


------------------------------------------------------
--- Typed Expressions
------------------------------------------------------

data TExp = 
           -- Constants
           TEUnit Type
         | TETrue Type
         | TEFalse Type
         | TENat Int Type

           -- Variables
         | TEVar VarId Type

           -- Functions
         | TEApp FunId TExp Type
         | TEUnaryOp UnaryOp TExp Type
         | TEBinOp BinOp TExp TExp Type

         -- Natural Numbers
         | TESucc TExp Type
         | TEToInt TExp Type
         | TEMatchN TExp TExp VarId TExp Type

           -- Let / Conditional
         | TELet Kind VarId TExp TExp Type
         | TECond TExp TExp TExp Type

           -- Lists
         | TENil Kind Type
         | TECons Kind TExp TExp Type
         | TEMatchL MatchType TExp TExp VarId VarId TExp Type

           -- Tuples
         | TETuple [TExp] Type
         | TEMatchT Kind TExp [VarId] TExp Type

           -- Binary Trees
         | TELeaf Kind Type
         | TENodeB Kind TExp TExp TExp Type
         | TEMatchB MatchType TExp TExp VarId VarId VarId TExp Type

           -- Ticks
         | TETick Double Type

           -- Sharing
         | TEShare VarId VarId VarId TExp Type  -- TEShare x1 x2 x e t means x ~> x1,x2

         deriving (Eq)



instance Show TExp where
    show e = showTabs e 0
     where
       showTabs expr tabs = case expr of
            TEUnit _ -> "()"

            TETrue _ -> "true"

            TEFalse _ -> "false"

            TENat n _ ->  show n

            TEApp f e _ -> f ++ " " ++ (show e)

            TEVar x _ -> x

            TEUnaryOp op e _ -> (show op) ++ " (" ++ (show e) ++ ")"

            TEBinOp op e1 e2 _ -> "(" ++ (show e1) ++ ")" ++ (show op) ++ "("++(show e2)++")"

            TESucc e _ -> "S " ++ (show e)

            TEToInt e _ -> "toInt " ++ (show e)

            TEMatchN e e1 x e2 _ ->
                          "match " ++ (show e) ++ " with" ++
                                   "\n" ++ (mkSpaces tabs) ++ "0 -> " ++ (showTabs e1 (tabs+7)) ++
                                   "\n" ++ (mkSpaces tabs) ++ "S("++x++") -> " 
                                        ++ (showTabs e2 (tabs+(length x)+8))

            TEShare x y z e _ -> "share " ++ x ++ "," ++ y ++ " = " ++ z ++" in" ++
                               "\n" ++ (mkSpaces tabs) ++ (showTabs e tabs)

            TELet Free x e1 e2 _ -> "freelet " ++ x ++ " = " ++ (showTabs e1 ((length x)+7) ) ++" in" ++
                                    "\n" ++ (mkSpaces tabs) ++ (showTabs e2 tabs)

            TELet Normal x e1 e2 _ -> "let " ++ x ++ " = " ++ (showTabs e1 ((length x)+7) ) ++" in" ++
                                      "\n" ++ (mkSpaces tabs) ++ (showTabs e2 tabs)

            TECond e e1 e2 _ -> "if " ++ (show e) ++ 
                               "\n"++(mkSpaces tabs)++"then "++(showTabs e1 (tabs+5))++
                               "\n"++(mkSpaces tabs)++"else "++(showTabs e2 (tabs+5))

            TENil Normal _ -> "nil"

            TENil Free _ -> "freenil"

            TECons Normal e1 e2 _ -> "cons(" ++ (show e1) ++ "," ++ (show e2) ++ ")"

            TECons Free e1 e2 _ -> "freecons(" ++ (show e1) ++ "," ++ (show e2) ++ ")"

            TEMatchL matchType e e1 x_h x_t e2 _ ->
                                let maybeD = case matchType of
                                               Dest -> "D"
                                               NonDest -> ""
                                in "match" ++ maybeD ++ " " ++ (show e) ++ " with" ++
                                   "\n" ++ (mkSpaces tabs) ++ "nil -> " ++ (showTabs e1 (tabs+7)) ++
                                   "\n" ++ (mkSpaces tabs) ++ "("++x_h++"::"++x_t++") -> " 
                                        ++ (showTabs e2 (tabs+(length (x_h++x_t)+8)))

            TETuple es _ -> "(" ++ (showWithSep "," es) ++")"

            TEMatchT _ e xs e1 _ -> let tuple = "("++ (showWithSep "," xs) ++") -> " in
                                   "match " ++ (show e) ++ " with" ++
                                   "\n" ++ (mkSpaces tabs) ++ tuple ++ (showTabs e1 (tabs+(length tuple)))

            TELeaf Normal _ -> "leaf"

            TELeaf Free _ -> "freeleaf"

            TENodeB Normal e1 e2 e3 _ -> "node(" ++ (show e1) ++ "," ++ (show e2) ++ "," ++ (show e3) ++ ")"

            TENodeB Free e1 e2 e3 _ -> "freenode(" ++ (show e1) ++ "," ++ (show e2) ++ "," ++ (show e3) ++ ")"

            TEMatchB matchType e e1 v t1 t2 e2 _ ->
                                let maybeD = case matchType of
                                               Dest -> "D"
                                               NonDest -> ""
                                in "match" ++ maybeD ++ " " ++ (show e) ++ " with" ++
                                   "\n" ++ (mkSpaces tabs) ++ "leaf -> " ++ (showTabs e1 (tabs+7)) ++
                                   "\n" ++ (mkSpaces tabs) ++ "node("++v++","++t1++","++t2++") -> " 
                                        ++ (showTabs e2 (tabs+(length (v++t1++t2)+10)))

            TETick q _ -> "tick(" ++ (show q) ++ ")"



typeOfTExp e = case e of
              TEUnit t -> t
              TETrue t -> t
              TEFalse t -> t
              TENat _ t -> t

              TEVar _ t -> t

              TEApp _ _ t -> t
              TEUnaryOp _ _ t -> t
              TEBinOp _ _ _ t -> t

              TESucc _ t -> t
              TEToInt _ t -> t
              TEMatchN _ _ _ _ t -> t

              TELet _ _ _ _ t -> t
              TECond _ _ _ t -> t

              TENil _ t -> t
              TECons _ _ _ t -> t
              TEMatchL _ _ _ _ _ _ t -> t

              TETuple _ t -> t
              TEMatchT _ _ _ _ t -> t

              TELeaf _ t -> t
              TENodeB _ _ _ _ t -> t
              TEMatchB _ _ _ _ _ _ _ t -> t

              TETick _ t -> t

              TEShare _ _ _ _ t -> t


instance Typed TExp where
    typeOf = typeOfTExp


freeVars :: TExp -> [VarId]

freeVars e = case e of
               TEUnit _ -> []
               TETrue _ -> []
               TEFalse _ -> []
               TENat _ _ -> []
               TEVar x _ -> [x]
               TEApp _ e _ -> freeVars e
               TEUnaryOp _ e _ -> freeVars e
               TEBinOp _ e1 e2 _ -> (freeVars e1) `union` (freeVars e2)
               TESucc e _ -> freeVars e
               TEToInt e _ -> freeVars e
               TEMatchN e e1 x e2 _ -> (freeVars e) `union` (freeVars e1) `union` ((freeVars e2) \\ [x])
               TELet _ x e1 e2 _ -> (freeVars e1) `union` ((freeVars e2) \\ [x])
               TEShare y y' x e _ -> [x] `union` ((freeVars e) \\ [y,y'])
               TECond e e1 e2 _ -> (freeVars e) `union` (freeVars e1) `union`  (freeVars e2)
               TENil _ _ -> []
               TECons _ e1 e2 _ -> (freeVars e1) `union` (freeVars e2)
               TEMatchL _ e e1 x1 x2 e2 _ -> (freeVars e) `union` (freeVars e1) `union` ((freeVars e2) \\ [x1,x2])
               TETuple es _ -> foldl union [] [freeVars e | e <- es]
               TEMatchT _ e xs e' _ -> (freeVars e) `union` ((freeVars e') \\ xs)
               TELeaf _ _ -> []
               TENodeB _ e1 e2 e3 _ -> (freeVars e1) `union` (freeVars e2) `union` (freeVars e3)
               TEMatchB _ e e1 v t1 t2 e2 _ -> (freeVars e) `union` (freeVars e1) `union` ((freeVars e2) \\ [v,t1,t2])
               TETick _ _ -> []





setType ex t = 
    case ex of
      TEUnit _ -> TEUnit t
      TETrue _ -> TETrue t
      TEFalse _ -> TEFalse t
      TENat n _ -> TENat n t

      TEVar x _ -> TEVar x t

      TEApp f e _ -> TEApp f e t

      TEUnaryOp UNot e _ -> TEUnaryOp UNot e t

      TEUnaryOp op e _ -> TEUnaryOp op (setType e t) t

      TEBinOp op e1 e2 _ | op `elem` [Add,Sub,Mult,Mod,Div,And,Or] ->
        TEBinOp op (setType e1 t) (setType e2 t) t

      TEBinOp op e1 e2 _ | op `elem` [Eq,LessEq,GreaterEq,Less,Greater] ->
        TEBinOp op e1 e2 t

      TESucc e _ -> TESucc (setType e t) t

      TEToInt e _ -> TEToInt e t

      TEMatchN e e1 x e2 _ -> TEMatchN e (setType e1 t) x (setType e2 t) t

      TELet kind x e1 e2 _ -> TELet kind x e1 (setType e2 t) t

      TECond e e1 e2 _ -> TECond e (setType e1 t) (setType e2 t) t

      TENil kind _  -> TENil kind t

      TECons kind e_h e_t _ -> 
          case t of
            TList t' -> TECons kind (setType e_h t') (setType e_t t) t
            _        -> error "Trying to assign a non-list type to a list."

      TEMatchL mt e e1 x_h x_t e2 _ -> 
          TEMatchL mt e (setType e1 t) x_h x_t (setType e2 t) t

      TETuple es _ -> 
          case t of
            TTuple ts -> TETuple (map2 setType es ts) t
            _         -> error "Trying to assign a non-tuple type to a tuple."

      TEMatchT mt e xs e' _ -> TEMatchT mt e xs (setType e' t) t
      TELeaf kind _ -> TELeaf kind t

      TENodeB kind e e1 e2 _ -> 
          case t of
            TBTree t' -> TENodeB kind (setType e t') (setType e1 t) (setType e2 t) t
            _         -> error "Trying to assign a non-tree type to a binary tree."

      TEMatchB mt e e1 x x1 x2 e2 _ -> TEMatchB mt e (setType e1 t) x x1 x2 (setType e2 t) t

      TEShare x y z e _ -> TEShare x y z (setType e t) t

      TETick q _ -> TETick q t



------------------------------------------------------
-- Parsed Objects
------------------------------------------------------

-- Everything that has a source position belongs to the type class
-- parsed.  That are expressions and declarations.
class Parsed a where 
   sourcePos :: a -> String

instance Parsed Exp where 
   sourcePos e = case e of
                   EUnit pos -> show pos
                   ETrue pos -> show pos
                   EFalse pos -> show pos
                   ENat _ pos -> show pos
                   EApp _ _ pos -> show pos
                   EVar _ pos -> show pos
                   EUnaryOp _ _ pos -> show pos
                   EBinOp _ _ _ pos -> show pos
                   ESucc _ pos -> show pos
                   EToInt _ pos -> show pos
                   EMatchN _ _ _ _ pos -> show pos         
                   ELet _ _ _ pos -> show pos
                   ECond _ _ _ pos -> show pos
                   ENil pos -> show pos
                   ECons _ _ pos -> show pos
                   EMatchL _ _ _ _ _ _ pos -> show pos
                   ETuple _ pos -> show pos
                   EMatchT _ _ _ pos -> show pos
                   ELeaf pos -> show pos
                   ENodeB _ _ _ pos -> show pos
                   EMatchB _ _ _ _ _ _ _ pos -> show pos
                   ETick _ pos -> show pos


instance Parsed Dec where
   sourcePos (TypeDec _ _ pos) = show pos
   sourcePos (FunDec _ _ _ pos) = show pos





------------------------------------------------------
-- Constraints
------------------------------------------------------

type CVarId = String


-- A constraint factor is ether a variable or a rational number.
-- CnVar is only needed for the change of the bases when sharing
-- variables.  That is why it uses Integer instead of rational.
data CFact = CnVar Integer CVarId | CVar CVarId | CConst Rational 
             deriving (Eq)

instance Show CFact where
   show (CnVar n x) = show n ++ " " ++ x
   show (CVar x) = x
   show (CConst q) = show (fromRational q) -- Unfortunately, lp_solve does only accept doubles
                                           -- so we have to convert rationals to doubles


-- We have three types of constraints.  The list of factors is read as
-- sum of factors.  E.g. CLeq [x,1,-4] [y,z,5] means x+1-4 <= y+z+5
data Constraint = CLeq [CFact] [CFact]
                | CGeq [CFact] [CFact]
                | CEq [CFact] [CFact]
                deriving (Eq)

instance Show Constraint where
   show constr = case constr of
                  CLeq sum1 sum2 -> (showWithSep " + " sum1) ++ " <= " ++ (showWithSep " + " sum2)
                  CGeq sum1 sum2 -> (showWithSep " + " sum1) ++ " >= " ++ (showWithSep " + " sum2)
                  CEq  sum1 sum2 -> (showWithSep " + " sum1) ++ " = " ++ (showWithSep " + " sum2)

type Constraints = [Constraint]



data Objective = Objective Rational CVarId

instance Show Objective where

   show (Objective r x) = (show (fromRational r)) ++ " " ++ x




------------------------------------------------------
-- Resource Metrics
------------------------------------------------------


data ResourceConstant = 
     KUnit
   | KNat
   | KBool
   | KVar
   | KUnary UnaryOp
   | KBin BinOp
   | KApp1 | KApp2
   | KSucc
   | KToInt
   | KMatchZ1 | KMatchZ2 | KMatchS1 | KMatchS2
   | KLet1 | KLet2 | KLet3
   | KConT1 | KConT2 | KConF1 | KConF2
   | KTuple Int
   | KMatchT1 Kind | KMatchT2 Kind
   | KNil
   | KCons Int
   | KMatchN1 MatchType | KMatchN2 MatchType | KMatchC1 MatchType Int | KMatchC2 MatchType
   | KLeaf
   | KNodeB Int
   | KMatchBL1 MatchType | KMatchBL2 MatchType | KMatchBN1 MatchType Int | KMatchBN2 MatchType
   | KTick Double


-- A resource metric consists of a name and a function that maps
-- resource constants to rational numbers.
data ResourceMetric = ResourceMetric String (ResourceConstant -> Rational)

instance Show ResourceMetric where
   show (ResourceMetric name f) = name

-- Equality on matrices is defined by comparing there names. 
instance Eq ResourceMetric where
   (==) (ResourceMetric str1 _) (ResourceMetric str2 _) = str1 == str2


-- The heap-space metric.  We store tuples on the stack.  If we create
-- a list then we put the tuples to the stack.  If we destroy a cons
-- cell the we move the tuple from the heap to the stack.
heapSpace = ResourceMetric "heap cells"
             (\k -> case k of 
                       KTuple n         -> 0
                       KCons n          -> fromIntegral $ 1 + n
                       KMatchC1 Dest n  -> fromIntegral $ -(1 + n)
                       KNodeB n         -> fromIntegral $ 2 + n
                       KMatchBN1 Dest n  -> fromIntegral $ -(2 + n)
                       _ -> 0
             )

costFree = ResourceMetric "cost-free" (\_ -> 0)


overflow = ResourceMetric "overflow" (\_ -> 0)


evalSteps = ResourceMetric "evaluation steps"
             (\k -> case k of 
                       KApp2 -> 0
                       KLet2 -> 0
                       KLet3 -> 0
                       KConT2 -> 0
                       KConF2 -> 0
                       KMatchZ2 -> 0
                       KMatchS2 -> 0
                       KMatchT1 Free -> 0
                       KMatchT2 _ -> 0
                       KMatchN2 _ -> 0
                       KMatchC2 _ -> 0
                       KMatchBL2 _ -> 0
                       KMatchBN2 _ -> 0
                       KUnary UPlus -> 0
                       KSucc -> 2 -- to make S n equivalent to n + 1
                       KTick _ -> 0
                       _ -> 1
             )

ticks = ResourceMetric "ticks"
           (\k -> case k of
                     KTick q -> toRational q
                     _ -> 0
           )



-- The following is needed to define infix operations in Typing.hs
-- to store inequalities easily.
-- What we wanna write is something like KLet1 + p <= q.
-- So we have to deal with ResourceConstants and Strings:

data SymbFact = SymbVar String 
                | SymbConst ResourceConstant

-- The operation (+) will be defined on lists of SumbFacts.  So
-- everything we use with (+) must implement the function makeSum that
-- converts an object to a list of SumbFacts.

class Summable a where
   makeSum :: a -> [SymbFact]

instance Summable ResourceConstant where
   makeSum r = [SymbConst r]

instance Summable CVarId where
   makeSum v = [SymbVar v]

instance Summable [SymbFact] where
   makeSum = id

-- A list of SumbFacts can then be converted to a sum of CFacts in
-- order to be used in constraints:

mkConstrSum (ResourceMetric _ metric) symbFacts = [ convert x | x <- symbFacts ]
       where
           convert (SymbVar id) = CVar id
           convert (SymbConst k) = CConst (metric k)





------------------------------------------------------
-- Values and Sizes
------------------------------------------------------


type Location = Integer

data Value = VLoc Location
           | VBool Bool
           | VInt Int
           | VNull
           | VList Value Value
           | VTuple [Value]
           | VBTree Value Value Value
  deriving (Eq,Show)



class Size a where 
   size :: a -> Int

instance Size Type where 
      size tt = case tt of
            TTuple ts -> foldl (\acc t -> acc + (size t)) 0 ts
            _         -> 1

instance Size Value where 
      size val = case val of
            VTuple vs -> foldl (\acc v -> acc + (size v)) 0 vs
            _         -> 1



------------------------------------------------------
-- Misc
------------------------------------------------------


data OutputMode = Verbose | Short
   deriving Eq


type Context = M.Map VarId Type

type FunDefs = M.Map FunId ([(VarId,Type)],TExp)



